import tempfile
from pathlib import Path

import pytest
from django.contrib.auth.models import AnonymousUser
from django.core.files.base import ContentFile

from umap.models import DataLayer, Map

from .base import DataLayerFactory, MapFactory

pytestmark = pytest.mark.django_db


def test_datalayers_should_be_ordered_by_rank(map, datalayer):
    datalayer.rank = 5
    datalayer.save()
    c4 = DataLayerFactory(map=map, rank=4)
    c1 = DataLayerFactory(map=map, rank=1)
    c3 = DataLayerFactory(map=map, rank=3)
    c2 = DataLayerFactory(map=map, rank=2)
    assert list(map.datalayer_set.all()) == [c1, c2, c3, c4, datalayer]


def test_upload_to(map, datalayer):
    map.pk = 302
    datalayer.pk = 17
    assert datalayer.upload_to().startswith("datalayer/2/0/302/17_")


def test_save_should_use_pk_as_name(map, datalayer):
    assert "/{}_".format(datalayer.pk) in datalayer.geojson.name


def test_same_geojson_file_name_will_be_suffixed(map, datalayer):
    before = datalayer.geojson.name
    datalayer.geojson.save(before, ContentFile("{}"))
    assert datalayer.geojson.name != before
    assert "/{}_".format(datalayer.pk) in datalayer.geojson.name


def test_clone_should_return_new_instance(map, datalayer):
    clone = datalayer.clone()
    assert datalayer.pk != clone.pk
    assert datalayer.name == clone.name
    assert datalayer.map == clone.map


def test_clone_should_update_map_if_passed(datalayer, user, licence):
    map = MapFactory(owner=user, licence=licence)
    clone = datalayer.clone(map_inst=map)
    assert datalayer.pk != clone.pk
    assert datalayer.name == clone.name
    assert datalayer.map != clone.map
    assert map == clone.map


def test_clone_should_clone_geojson_too(datalayer):
    clone = datalayer.clone()
    assert datalayer.pk != clone.pk
    assert clone.geojson is not None
    assert clone.geojson.path != datalayer.geojson.path


def test_should_remove_old_versions_on_save(map, settings):
    datalayer = DataLayerFactory(uuid="0f1161c0-c07f-4ba4-86c5-8d8981d8a813", old_id=17)
    settings.UMAP_KEEP_VERSIONS = 3
    root = Path(datalayer.storage_root())
    before = len(datalayer.geojson.storage.listdir(root)[1])
    newer = f"{datalayer.pk}_1440924889.geojson"
    medium = f"{datalayer.pk}_1440923687.geojson"
    older = f"{datalayer.pk}_1440918637.geojson"
    with_old_id = f"{datalayer.old_id}_1440918537.geojson"
    other = "123456_1440918637.geojson"
    for path in [medium, newer, older, with_old_id, other]:
        datalayer.geojson.storage.save(root / path, ContentFile("{}"))
        datalayer.geojson.storage.save(root / f"{path}.gz", ContentFile("{}"))
    assert len(datalayer.geojson.storage.listdir(root)[1]) == 10 + before
    files = datalayer.geojson.storage.listdir(root)[1]
    # Those files should be present before save, which will purge them
    assert older in files
    assert older + ".gz" in files
    assert with_old_id in files
    assert with_old_id + ".gz" in files
    datalayer.save()
    files = datalayer.geojson.storage.listdir(root)[1]
    # Flat + gz files, but not latest gz, which is created at first datalayer read.
    # older and with_old_id should have been removed
    assert len(files) == 5
    assert newer in files
    assert medium in files
    assert Path(datalayer.geojson.path).name in files
    # File from another datalayer, purge should have impacted it.
    assert other in files
    assert other + ".gz" in files
    assert older not in files
    assert older + ".gz" not in files
    assert with_old_id not in files
    assert with_old_id + ".gz" not in files
    names = [v["name"] for v in datalayer.versions]
    assert names == [Path(datalayer.geojson.name).name, newer, medium]


def test_anonymous_cannot_edit_in_editors_mode(datalayer, fake_request):
    datalayer.edit_status = DataLayer.COLLABORATORS
    datalayer.save()
    fake_request.user = AnonymousUser()
    assert not datalayer.can_edit(fake_request)


def test_owner_can_edit_in_editors_mode(datalayer, user, fake_request):
    datalayer.edit_status = DataLayer.COLLABORATORS
    datalayer.save()
    fake_request.user = datalayer.map.owner
    assert datalayer.can_edit(fake_request)


def test_editor_can_edit_in_collaborators_mode(datalayer, user, fake_request):
    map = datalayer.map
    map.editors.add(user)
    map.save()
    datalayer.edit_status = DataLayer.COLLABORATORS
    datalayer.save()
    fake_request.user = user
    assert datalayer.can_edit(fake_request)


def test_team_members_can_edit_in_collaborators_mode(
    datalayer, user, team, fake_request
):
    user.teams.add(team)
    user.save()
    map = datalayer.map
    map.team = team
    map.save()
    datalayer.edit_status = DataLayer.COLLABORATORS
    datalayer.save()
    fake_request.user = user
    assert datalayer.can_edit(fake_request)


def test_anonymous_can_edit_in_public_mode(datalayer, fake_request):
    datalayer.edit_status = DataLayer.ANONYMOUS
    datalayer.save()
    fake_request.user = AnonymousUser()
    assert datalayer.can_edit(fake_request)


def test_owner_can_edit_in_public_mode(datalayer, user, fake_request):
    datalayer.edit_status = DataLayer.ANONYMOUS
    datalayer.save()
    fake_request.user = datalayer.map.owner
    assert datalayer.can_edit(fake_request)


def test_editor_can_edit_in_public_mode(datalayer, user, fake_request):
    map = datalayer.map
    map.editors.add(user)
    map.save()
    datalayer.edit_status = DataLayer.ANONYMOUS
    datalayer.save()
    fake_request.user = user
    assert datalayer.can_edit(fake_request)


def test_anonymous_cannot_edit_in_anonymous_owner_mode(datalayer, fake_request):
    datalayer.edit_status = DataLayer.OWNER
    datalayer.save()
    map = datalayer.map
    map.owner = None
    map.save()
    fake_request.user = AnonymousUser()
    assert not datalayer.can_edit(fake_request)


def test_owner_can_edit_in_inherit_mode_and_map_in_owner_mode(datalayer, fake_request):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.edit_status = Map.OWNER
    map.save()
    fake_request.user = map.owner
    assert datalayer.can_edit(fake_request)


def test_editors_cannot_edit_in_inherit_mode_and_map_in_owner_mode(
    datalayer, user, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.editors.add(user)
    map.edit_status = Map.OWNER
    map.save()
    fake_request.user = user
    assert not datalayer.can_edit(fake_request)


def test_team_members_cannot_edit_in_inherit_mode_and_map_in_owner_mode(
    datalayer, user, team, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    user.teams.add(team)
    team.save()
    map = datalayer.map
    map.team = team
    map.edit_status = Map.OWNER
    map.save()
    fake_request.user = user
    assert not datalayer.can_edit(fake_request)


def test_anonymous_cannot_edit_in_inherit_mode_and_map_in_owner_mode(
    datalayer, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.edit_status = Map.OWNER
    map.save()
    fake_request.user = AnonymousUser()
    assert not datalayer.can_edit(fake_request)


def test_owner_can_edit_in_inherit_mode_and_map_in_editors_mode(
    datalayer, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.edit_status = Map.COLLABORATORS
    map.save()
    fake_request.user = map.owner
    assert datalayer.can_edit(fake_request)


def test_editors_can_edit_in_inherit_mode_and_map_in_editors_mode(
    datalayer, user, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.editors.add(user)
    map.edit_status = Map.COLLABORATORS
    map.save()
    fake_request.user = user
    assert datalayer.can_edit(fake_request)


def test_anonymous_cannot_edit_in_inherit_mode_and_map_in_editors_mode(
    datalayer, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.edit_status = Map.COLLABORATORS
    map.save()
    fake_request.user = AnonymousUser()
    assert not datalayer.can_edit(fake_request)


def test_anonymous_can_edit_in_inherit_mode_and_map_in_public_mode(
    datalayer, fake_request
):
    datalayer.edit_status = DataLayer.INHERIT
    datalayer.save()
    map = datalayer.map
    map.edit_status = Map.ANONYMOUS
    map.save()
    fake_request.user = AnonymousUser()
    assert datalayer.can_edit(fake_request)


def test_should_remove_all_versions_on_delete(map, settings):
    settings.UMAP_PURGATORY_ROOT = tempfile.mkdtemp()
    datalayer = DataLayerFactory(uuid="0f1161c0-c07f-4ba4-86c5-8d8981d8a813", old_id=17)
    root = Path(datalayer.storage_root())
    before = len(datalayer.geojson.storage.listdir(root)[1])
    other = "123456_1440918637.geojson"
    files = [
        f"{datalayer.pk}_1440924889.geojson",
        f"{datalayer.pk}_1440923687.geojson",
        f"{datalayer.pk}_1440918637.geojson",
        f"{datalayer.old_id}_1440918537.geojson",
        other,
    ]
    for path in files:
        datalayer.geojson.storage.save(root / path, ContentFile("{}"))
        datalayer.geojson.storage.save(root / f"{path}.gz", ContentFile("{}"))
    assert len(datalayer.geojson.storage.listdir(root)[1]) == 10 + before
    datalayer.delete()
    found = datalayer.geojson.storage.listdir(root)[1]
    assert found == [other, f"{other}.gz"]
    assert len(list(Path(settings.UMAP_PURGATORY_ROOT).iterdir())) == 4 + before



import pytest
from playwright.sync_api import expect

from ..base import DataLayerFactory

pytestmark = pytest.mark.django_db

DATALAYER_DATA = {
    "type": "FeatureCollection",
    "features": [
        {
            "type": "Feature",
            "properties": {"name": "name line", "description": "line description"},
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    # Flat line so PW will click on it
                    # (it compute the center of the element)
                    [11.25, 53.585984],
                    [10.151367, 52.975108],
                ],
            },
        },
    ],
}


@pytest.fixture
def bootstrap(map, live_server):
    map.settings["properties"]["zoom"] = 6
    map.settings["geometry"] = {
        "type": "Point",
        "coordinates": [8.429, 53.239],
    }
    map.save()
    DataLayerFactory(map=map, data=DATALAYER_DATA)


def test_should_open_popup_on_click(live_server, map, page, bootstrap):
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    line = page.locator("path").first
    expect(line).to_have_attribute("stroke-opacity", "0.5")
    line.click()
    expect(page.locator(".leaflet-popup-content-wrapper")).to_be_visible()
    expect(page.get_by_role("heading", name="name line")).to_be_visible()
    expect(page.get_by_text("line description")).to_be_visible()
    # It's not a round value
    expect(line).to_have_attribute("stroke-opacity", "1")
    # Close popup
    page.locator("#map").click()
    expect(line).to_have_attribute("stroke-opacity", "0.5")



import json
import operator
import os
import shutil
import time
import uuid
from pathlib import Path

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.gis.db import models
from django.core.files.base import File
from django.core.signing import Signer
from django.template.defaultfilters import slugify
from django.urls import reverse
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from .managers import PublicManager
from .utils import _urls_for_js


# Did not find a clean way to do this in Django
# - creating a Proxy model would mean replacing get_user_model by this proxy model
#   in every template
# - extending User model would mean a non trivial migration
def display_name(self):
    return settings.USER_DISPLAY_NAME.format(**self.__dict__)


def get_user_url(self):
    identifier = getattr(self, settings.USER_URL_FIELD)
    return reverse(settings.USER_MAPS_URL, kwargs={"identifier": identifier})


def get_user_stars_url(self):
    identifier = getattr(self, settings.USER_URL_FIELD)
    return reverse("user_stars", kwargs={"identifier": identifier})


User.add_to_class("__str__", display_name)
User.add_to_class("get_url", get_user_url)
User.add_to_class("get_stars_url", get_user_stars_url)


def get_default_share_status():
    return settings.UMAP_DEFAULT_SHARE_STATUS or Map.PUBLIC


def get_default_edit_status():
    return settings.UMAP_DEFAULT_EDIT_STATUS or Map.OWNER


class Team(models.Model):
    name = models.CharField(
        max_length=200, verbose_name=_("name"), unique=True, blank=False, null=False
    )
    description = models.TextField(blank=True, null=True, verbose_name=_("description"))
    users = models.ManyToManyField(User, related_name="teams")

    def __unicode__(self):
        return self.name

    def __str__(self):
        return self.name

    def get_url(self):
        return reverse("team_maps", kwargs={"pk": self.pk})

    def get_metadata(self):
        return {"id": self.pk, "name": self.name, "url": self.get_url()}


class NamedModel(models.Model):
    name = models.CharField(max_length=200, verbose_name=_("name"))

    class Meta:
        abstract = True
        ordering = ("name",)

    def __unicode__(self):
        return self.name

    def __str__(self):
        return self.name


def get_default_licence():
    """
    Returns a default Licence, creates it if it doesn't exist.
    Needed to prevent a licence deletion from deleting all the linked
    maps.
    """
    return Licence.objects.get_or_create(
        # can't use ugettext_lazy for database storage, see #13965
        name=getattr(settings, "UMAP_DEFAULT_LICENCE_NAME", "No licence set")
    )[0]


class Licence(NamedModel):
    """
    The licence one map is published on.
    """

    details = models.URLField(
        verbose_name=_("details"),
        help_text=_("Link to a page where the licence is detailed."),
    )

    @property
    def json(self):
        return {"name": self.name, "url": self.details}


class TileLayer(NamedModel):
    url_template = models.CharField(
        max_length=400, help_text=_("URL template using OSM tile format")
    )
    minZoom = models.IntegerField(default=0)
    maxZoom = models.IntegerField(default=18)
    attribution = models.CharField(max_length=300)
    rank = models.SmallIntegerField(
        blank=True, null=True, help_text=_("Order of the tilelayers in the edit box")
    )
    # See https://wiki.openstreetmap.org/wiki/TMS#The_Y_coordinate
    tms = models.BooleanField(default=False)

    @property
    def json(self):
        return dict(
            (field.name, getattr(self, field.name)) for field in self._meta.fields
        )

    @classmethod
    def get_default(cls):
        """
        Returns the default tile layer (used for a map when no layer is set).
        """
        return cls.objects.order_by("rank")[0]  # FIXME, make it administrable

    @classmethod
    def get_list(cls):
        default = cls.get_default()
        l = []
        for t in cls.objects.all():
            fields = t.json
            if default and default.pk == t.pk:
                fields["selected"] = True
            l.append(fields)
        return l

    class Meta:
        ordering = ("rank", "name")


class Map(NamedModel):
    """
    A single thematical map.
    """

    ANONYMOUS = 1
    COLLABORATORS = 2
    OWNER = 3
    PUBLIC = 1
    OPEN = 2
    PRIVATE = 3
    BLOCKED = 9
    EDIT_STATUS = (
        (ANONYMOUS, _("Everyone")),
        (COLLABORATORS, _("Editors and team only")),
        (OWNER, _("Owner only")),
    )
    SHARE_STATUS = (
        (PUBLIC, _("Everyone (public)")),
        (OPEN, _("Anyone with link")),
        (PRIVATE, _("Editors and team only")),
        (BLOCKED, _("Blocked")),
    )
    slug = models.SlugField(db_index=True)
    center = models.PointField(geography=True, verbose_name=_("center"))
    zoom = models.IntegerField(default=7, verbose_name=_("zoom"))
    locate = models.BooleanField(
        default=False, verbose_name=_("locate"), help_text=_("Locate user on load?")
    )
    licence = models.ForeignKey(
        Licence,
        help_text=_("Choose the map licence."),
        verbose_name=_("licence"),
        on_delete=models.SET_DEFAULT,
        default=get_default_licence,
    )
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        related_name="owned_maps",
        verbose_name=_("owner"),
        on_delete=models.PROTECT,
    )
    editors = models.ManyToManyField(
        settings.AUTH_USER_MODEL, blank=True, verbose_name=_("editors")
    )
    team = models.ForeignKey(
        Team,
        blank=True,
        null=True,
        verbose_name=_("team"),
        on_delete=models.SET_NULL,
    )
    edit_status = models.SmallIntegerField(
        choices=EDIT_STATUS,
        default=get_default_edit_status,
        verbose_name=_("edit status"),
    )
    share_status = models.SmallIntegerField(
        choices=SHARE_STATUS,
        default=get_default_share_status,
        verbose_name=_("share status"),
    )
    settings = models.JSONField(
        blank=True, null=True, verbose_name=_("settings"), default=dict
    )

    objects = models.Manager()
    public = PublicManager()

    @property
    def description(self):
        try:
            return self.settings["properties"]["description"]
        except KeyError:
            return ""

    @property
    def preview_settings(self):
        layers = self.datalayer_set.all()
        datalayer_data = [c.metadata() for c in layers]
        map_settings = self.settings
        if "properties" not in map_settings:
            map_settings["properties"] = {}
        map_settings["properties"].update(
            {
                "tilelayers": [TileLayer.get_default().json],
                "datalayers": datalayer_data,
                "urls": _urls_for_js(),
                "STATIC_URL": settings.STATIC_URL,
                "editMode": "disabled",
                "hash": False,
                "scrollWheelZoom": False,
                "noControl": True,
                "umap_id": self.pk,
                "schema": self.extra_schema,
                "slideshow": {},
            }
        )
        return map_settings

    def delete(self, **kwargs):
        # Explicitely call datalayers.delete, so we can deal with removing files
        # (the cascade delete would not call the model delete method)
        for datalayer in self.datalayer_set.all():
            datalayer.delete()
        return super().delete(**kwargs)

    def generate_umapjson(self, request):
        umapjson = self.settings
        umapjson["type"] = "umap"
        umapjson["uri"] = request.build_absolute_uri(self.get_absolute_url())
        datalayers = []
        for datalayer in self.datalayer_set.all():
            with open(datalayer.geojson.path, "rb") as f:
                layer = json.loads(f.read())
            if datalayer.settings:
                layer["_umap_options"] = datalayer.settings
            datalayers.append(layer)
        umapjson["layers"] = datalayers
        return umapjson

    def get_absolute_url(self):
        return reverse("map", kwargs={"slug": self.slug or "map", "map_id": self.pk})

    def get_anonymous_edit_url(self):
        signer = Signer()
        signature = signer.sign(self.pk)
        path = reverse("map_anonymous_edit_url", kwargs={"signature": signature})
        return settings.SITE_URL + path

    def get_author(self):
        return self.team or self.owner

    def is_owner(self, request=None):
        if not request:
            return False
        if request.user and self.owner == request.user:
            return True
        return self.is_anonymous_owner(request)

    def is_anonymous_owner(self, request):
        if not request or self.owner:
            # edit cookies are only valid while the map doesn't have owner
            return False
        key, value = self.signed_cookie_elements
        try:
            has_anonymous_cookie = int(request.get_signed_cookie(key, False)) == value
        except ValueError:
            has_anonymous_cookie = False
        return has_anonymous_cookie

    def can_delete(self, request=None):
        if not request:
            return False
        if self.owner and request.user != self.owner:
            return False
        if not self.owner and not self.is_anonymous_owner(request):
            return False
        return True

    def can_edit(self, request=None):
        """
        Define if a user can edit or not the instance, according to his account
        or the request.

        In owner mode:
            - only owner by default (OWNER)
            - any editor or team member if mode is COLLABORATORS
            - anyone otherwise (ANONYMOUS)
        In anonymous owner mode:
            - only owner (has ownership cookie) by default (OWNER)
            - anyone otherwise (ANONYMOUS)
        """
        can = False
        if not request:
            return False
        user = request.user
        if (
            not self.owner
            and settings.UMAP_ALLOW_ANONYMOUS
            and self.is_anonymous_owner(request)
        ):
            can = True
        elif self.edit_status == self.ANONYMOUS:
            can = True
        elif not user.is_authenticated:
            can = False
        elif user == self.owner:
            can = True
        elif self.edit_status == self.COLLABORATORS:
            if user in self.editors.all() or self.team in user.teams.all():
                can = True
        return can

    def can_view(self, request):
        if self.share_status == self.BLOCKED:
            can = False
        elif self.owner is None:
            can = True
        elif self.share_status in [self.PUBLIC, self.OPEN]:
            can = True
        elif not request.user.is_authenticated:
            can = False
        elif request.user == self.owner:
            can = True
        else:
            can = not (
                self.share_status == self.PRIVATE
                and request.user not in self.editors.all()
                and self.team not in request.user.teams.all()
            )
        return can

    @property
    def signed_cookie_elements(self):
        return ("anonymous_owner|%s" % self.pk, self.pk)

    def get_tilelayer(self):
        return self.tilelayer or TileLayer.get_default()

    def clone(self, **kwargs):
        new = self.__class__.objects.get(pk=self.pk)
        new.pk = None
        new.name = "%s %s" % (_("Clone of"), self.name)
        if "owner" in kwargs:
            # can be None in case of anonymous cloning
            new.owner = kwargs["owner"]
        new.save()
        for editor in self.editors.all():
            new.editors.add(editor)
        for datalayer in self.datalayer_set.all():
            datalayer.clone(map_inst=new)
        return new

    @classproperty
    def extra_schema(self):
        return {
            "iconUrl": {
                "default": "%sumap/img/marker.svg" % settings.STATIC_URL,
            }
        }


class Pictogram(NamedModel):
    """
    An image added to an icon of the map.
    """

    attribution = models.CharField(max_length=300)
    category = models.CharField(max_length=300, null=True, blank=True)
    pictogram = models.FileField(upload_to="pictogram")

    @property
    def json(self):
        return {
            "id": self.pk,
            "attribution": self.attribution,
            "name": self.name,
            "category": self.category,
            "src": self.pictogram.url,
        }


# Must be out of Datalayer for Django migration to run, because of python 2
# serialize limitations.
def upload_to(instance, filename):
    if instance.pk:
        return instance.upload_to()
    name = "%s.geojson" % slugify(instance.name)[:50] or "untitled"
    return os.path.join(instance.storage_root(), name)


class DataLayer(NamedModel):
    """
    Layer to store Features in.
    """

    INHERIT = 0
    ANONYMOUS = 1
    COLLABORATORS = 2
    OWNER = 3
    EDIT_STATUS = (
        (INHERIT, _("Inherit")),
        (ANONYMOUS, _("Everyone")),
        (COLLABORATORS, _("Editors and team only")),
        (OWNER, _("Owner only")),
    )
    uuid = models.UUIDField(
        unique=True, primary_key=True, default=uuid.uuid4, editable=False
    )
    old_id = models.IntegerField(null=True, blank=True)
    map = models.ForeignKey(Map, on_delete=models.CASCADE)
    description = models.TextField(blank=True, null=True, verbose_name=_("description"))
    geojson = models.FileField(upload_to=upload_to, blank=True, null=True)
    display_on_load = models.BooleanField(
        default=False,
        verbose_name=_("display on load"),
        help_text=_("Display this layer on load."),
    )
    rank = models.SmallIntegerField(default=0)
    settings = models.JSONField(
        blank=True, null=True, verbose_name=_("settings"), default=dict
    )
    edit_status = models.SmallIntegerField(
        choices=EDIT_STATUS,
        default=INHERIT,
        verbose_name=_("edit status"),
    )

    class Meta:
        ordering = ("rank",)

    def save(self, force_insert=False, force_update=False, **kwargs):
        is_new = not bool(self.pk)
        super(DataLayer, self).save(
            force_insert=force_insert, force_update=force_update, **kwargs
        )

        if is_new:
            force_insert, force_update = False, True
            filename = self.upload_to()
            old_name = self.geojson.name
            new_name = self.geojson.storage.save(filename, self.geojson)
            self.geojson.storage.delete(old_name)
            self.geojson.name = new_name
            super(DataLayer, self).save(
                force_insert=force_insert, force_update=force_update, **kwargs
            )
        self.purge_gzip()
        self.purge_old_versions()

    def delete(self, **kwargs):
        self.purge_gzip()
        self.to_purgatory()
        return super().delete(**kwargs)

    def to_purgatory(self):
        dest = Path(settings.UMAP_PURGATORY_ROOT)
        dest.mkdir(parents=True, exist_ok=True)
        src = Path(self.geojson.storage.location) / self.storage_root()
        for version in self.versions:
            name = version["name"]
            shutil.move(src / name, dest / f"{self.map.pk}_{name}")

    def upload_to(self):
        root = self.storage_root()
        name = "%s_%s.geojson" % (self.pk, int(time.time() * 1000))
        return os.path.join(root, name)

    def storage_root(self):
        path = ["datalayer", str(self.map.pk)[-1]]
        if len(str(self.map.pk)) > 1:
            path.append(str(self.map.pk)[-2])
        path.append(str(self.map.pk))
        return os.path.join(*path)

    def metadata(self, request=None):
        # Retrocompat: minimal settings for maps not saved after settings property
        # has been introduced
        obj = self.settings or {
            "name": self.name,
            "displayOnLoad": self.display_on_load,
        }
        if self.old_id:
            obj["old_id"] = self.old_id
        obj["id"] = self.pk
        obj["permissions"] = {"edit_status": self.edit_status}
        obj["editMode"] = "advanced" if self.can_edit(request) else "disabled"
        return obj

    def clone(self, map_inst=None):
        new = self.__class__.objects.get(pk=self.pk)
        new._state.adding = True
        new.pk = None
        if map_inst:
            new.map = map_inst
        new.geojson = File(new.geojson.file.file)
        new.save()
        return new

    def is_valid_version(self, name):
        valid_prefixes = [name.startswith("%s_" % self.pk)]
        if self.old_id:
            valid_prefixes.append(name.startswith("%s_" % self.old_id))
        return any(valid_prefixes) and name.endswith(".geojson")

    def version_metadata(self, name):
        els = name.split(".")[0].split("_")
        return {
            "name": name,
            "at": els[1],
            "size": self.geojson.storage.size(self.get_version_path(name)),
        }

    @property
    def versions(self):
        root = self.storage_root()
        names = self.geojson.storage.listdir(root)[1]
        names = [name for name in names if self.is_valid_version(name)]
        versions = [self.version_metadata(name) for name in names]
        versions.sort(reverse=True, key=operator.itemgetter("at"))
        return versions

    def get_version(self, name):
        path = self.get_version_path(name)
        with self.geojson.storage.open(path, "r") as f:
            return f.read()

    def get_version_path(self, name):
        return "{root}/{name}".format(root=self.storage_root(), name=name)

    def purge_old_versions(self):
        root = self.storage_root()
        versions = self.versions[settings.UMAP_KEEP_VERSIONS :]
        for version in versions:
            name = version["name"]
            # Should not be in the list, but ensure to not delete the file
            # currently used in database
            if self.geojson.name.endswith(name):
                continue
            try:
                self.geojson.storage.delete(os.path.join(root, name))
            except FileNotFoundError:
                pass

    def purge_gzip(self):
        root = self.storage_root()
        names = self.geojson.storage.listdir(root)[1]
        prefixes = [f"{self.pk}_"]
        if self.old_id:
            prefixes.append(f"{self.old_id}_")
        prefixes = tuple(prefixes)
        for name in names:
            if name.startswith(prefixes) and name.endswith(".gz"):
                self.geojson.storage.delete(os.path.join(root, name))

    def can_edit(self, request=None):
        """
        Define if a user can edit or not the instance, according to his account
        or the request.
        """
        if self.edit_status == self.INHERIT:
            return self.map.can_edit(request)
        can = False
        if not request:
            return False
        user = request.user
        if not self.map.owner:
            if settings.UMAP_ALLOW_ANONYMOUS and self.map.is_anonymous_owner(request):
                can = True
        if self.edit_status == self.ANONYMOUS:
            can = True
        elif user.is_authenticated and user == self.map.owner:
            can = True
        elif user.is_authenticated and self.edit_status == self.COLLABORATORS:
            if user in self.map.editors.all() or self.map.team in user.teams.all():
                can = True
        return can


class Star(models.Model):
    at = models.DateTimeField(auto_now=True)
    map = models.ForeignKey(Map, on_delete=models.CASCADE)
    by = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="stars", on_delete=models.CASCADE
    )



# Generated by Django 4.2 on 2024-02-26 14:09

import re

from django.conf import settings
from django.db import migrations
from django.urls import NoReverseMatch, reverse

# Some users hacked uMap to use another map datalayer as a remote data source.
# This script gently handles the migration for them.


def migrate_datalayers(apps, schema_editor):
    DataLayer = apps.get_model("umap", "DataLayer")

    datalayers = DataLayer.objects.filter(
        settings__remoteData__url__icontains="datalayer"
    )

    for item in datalayers:
        old_url = item.settings["remoteData"]["url"]
        match = re.search(
            rf"{settings.SITE_URL}/datalayer/(?P<map_id>\d+)/(?P<datalayer_id>\d+)",
            old_url,
        )
        if match:
            remote_id = match.group("datalayer_id")
            map_id = match.group("map_id")
            try:
                remote_uuid = DataLayer.objects.get(old_id=remote_id).uuid
            except DataLayer.DoesNotExist:
                pass
            else:
                try:
                    new_url = settings.SITE_URL + reverse(
                        "datalayer_view", args=[map_id, remote_uuid]
                    )
                except NoReverseMatch:
                    pass
                else:
                    item.settings["remoteData"]["url"] = new_url
                    item.save()


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0018_datalayer_uuid"),
    ]

    operations = [
        migrations.RunPython(migrate_datalayers, reverse_code=migrations.RunPython.noop)
    ]



import json
import re
from pathlib import Path

from playwright.sync_api import expect

from umap.models import DataLayer

from ..base import DataLayerFactory

DATALAYER_DATA = {
    "type": "FeatureCollection",
    "features": [
        {
            "type": "Feature",
            "properties": {
                "mytype": "even",
                "name": "Point 2",
                "mynumber": 10,
                "myboolean": True,
                "mydate": "2024/04/14 12:19:17",
            },
            "geometry": {"type": "Point", "coordinates": [0.065918, 48.385442]},
            "id": "poin2",  # Must be exactly 5 chars long so the frontend will keep it
        },
        {
            "type": "Feature",
            "properties": {
                "mytype": "odd",
                "name": "Point 1",
                "mynumber": 12,
                "myboolean": False,
                "mydate": "2024/03/13 12:20:20",
            },
            "geometry": {"type": "Point", "coordinates": [3.55957, 49.767074]},
            "id": "poin1",
        },
        {
            "type": "Feature",
            "properties": {
                "mytype": "even",
                "name": "Point 4",
                "mynumber": 10,
                "myboolean": "true",
                "mydate": "2024/08/18 13:14:15",
            },
            "geometry": {"type": "Point", "coordinates": [0.856934, 45.290347]},
            "id": "poin4",
        },
        {
            "type": "Feature",
            "properties": {
                "mytype": "odd",
                "name": "Point 3",
                "mynumber": 14,
                "mydate": "2024-04-14T10:19:17.000Z",
            },
            "geometry": {"type": "Point", "coordinates": [4.372559, 47.945786]},
            "id": "poin3",
        },
    ],
    "_umap_options": {
        "name": "Calque 2",
    },
}


def test_table_editor(live_server, openmap, datalayer, page):
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    page.get_by_text("Add a new property").click()
    page.locator("dialog").locator("input").fill("newprop")
    page.locator("dialog").get_by_role("button", name="OK").click()
    page.locator("td").nth(2).dblclick()
    page.locator('input[name="newprop"]').fill("newvalue")
    page.keyboard.press("Enter")
    page.locator("thead button[data-property=name]").click()
    page.get_by_role("button", name="Delete this column").click()
    page.locator("dialog").get_by_role("button", name="OK").click()
    with page.expect_response(re.compile(r".*/datalayer/update/.*")):
        page.get_by_role("button", name="Save").click()
    saved = DataLayer.objects.last()
    data = json.loads(Path(saved.geojson.path).read_text())
    assert data["features"][0]["properties"]["newprop"] == "newvalue"
    assert "name" not in data["features"][0]["properties"]


def test_cannot_add_existing_property_name(live_server, openmap, datalayer, page):
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    page.get_by_text("Add a new property").click()
    page.locator("dialog").locator("input").fill("name")
    page.get_by_role("button", name="OK").click()
    expect(page.get_by_role("dialog")).to_contain_text(
        "This name already exists: “name”"
    )
    expect(page.locator("table th button[data-property=name]")).to_have_count(1)


def test_cannot_add_property_with_a_dot(live_server, openmap, datalayer, page):
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    page.get_by_text("Add a new property").click()
    page.locator("dialog").locator("input").fill("foo.bar")
    page.get_by_role("button", name="OK").click()
    expect(page.get_by_role("dialog")).to_contain_text(
        "Name “foo.bar” should not contain a dot."
    )
    expect(page.locator("table th button[data-property=name]")).to_have_count(1)


def test_rename_property(live_server, openmap, datalayer, page):
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    expect(page.locator("table th button[data-property=name]")).to_have_count(1)
    page.locator("thead button[data-property=name]").click()
    page.get_by_text("Rename this column").click()
    page.locator("dialog").locator("input").fill("newname")
    page.get_by_role("button", name="OK").click()
    expect(page.locator("table th button[data-property=newname]")).to_have_count(1)
    expect(page.locator("table th button[data-property=name]")).to_have_count(0)


def test_delete_selected_rows(live_server, openmap, page):
    DataLayerFactory(map=openmap, data=DATALAYER_DATA)
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit#6/48.093/1.890")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    expect(page.locator("tbody tr")).to_have_count(4)
    expect(page.locator(".leaflet-marker-icon")).to_have_count(4)
    page.locator("tr[data-feature=poin2]").get_by_role("checkbox").check()
    page.get_by_role("button", name="Delete selected rows").click()
    page.get_by_role("button", name="OK").click()
    expect(page.locator("tbody tr")).to_have_count(3)
    expect(page.locator(".leaflet-marker-icon")).to_have_count(3)


def test_delete_all_rows(live_server, openmap, page):
    DataLayerFactory(map=openmap, data=DATALAYER_DATA)
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit#6/48.093/1.890")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    expect(page.locator("tbody tr")).to_have_count(4)
    expect(page.locator(".leaflet-marker-icon")).to_have_count(4)
    page.locator("thead").get_by_role("checkbox").check()
    page.get_by_role("button", name="Delete selected rows").click()
    page.get_by_role("button", name="OK").click()
    expect(page.locator("tbody tr")).to_have_count(0)
    expect(page.locator(".leaflet-marker-icon")).to_have_count(0)


def test_filter_and_delete_rows(live_server, openmap, page):
    DataLayerFactory(map=openmap, data=DATALAYER_DATA)
    panel = page.locator(".panel.left.on")
    table = page.locator(".panel.full table")
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit#6/48.093/1.890")
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit properties in a table").click()
    expect(table.locator("tbody tr")).to_have_count(4)
    expect(page.locator(".leaflet-marker-icon")).to_have_count(4)
    table.locator("thead button[data-property=mytype]").click()
    page.get_by_role("button", name="Add filter for this column").click()
    expect(panel).to_be_visible()
    panel.get_by_label("even").check()
    table.locator("thead").get_by_role("checkbox").check()
    page.get_by_role("button", name="Delete selected rows").click()
    page.get_by_role("button", name="OK").click()
    expect(table.locator("tbody tr")).to_have_count(2)
    expect(page.locator(".leaflet-marker-icon")).to_have_count(2)
    expect(table.get_by_text("Point 1")).to_be_visible()
    expect(table.get_by_text("Point 3")).to_be_visible()
    expect(table.get_by_text("Point 2")).to_be_hidden()
    expect(table.get_by_text("Point 4")).to_be_hidden()



# Generated by Django 1.10.3 on 2016-11-26 16:02
from __future__ import unicode_literals

from django.db import migrations


def add_tilelayer(apps, *args):
    TileLayer = apps.get_model("umap", "TileLayer")
    if TileLayer.objects.count():
        return
    TileLayer(
        name="Positron",
        url_template=(
            "https://cartodb-basemaps-{s}.global.ssl.fastly.net/"
            "light_all/{z}/{x}/{y}.png"
        ),
        attribution=(
            "&copy; [[http://www.openstreetmap.org/copyright|"
            "OpenStreetMap]] contributors, &copy; "
            "[[https://carto.com/attributions|CARTO]]"
        ),
    ).save()


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0002_tilelayer_tms"),
    ]

    operations = [
        migrations.RunPython(add_tilelayer),
    ]



# Generated by Django 2.0.7 on 2018-07-14 11:58

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0004_add_licence"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="map",
            name="tilelayer",
        ),
    ]



import os
import tempfile
from pathlib import Path

from django.core.management import call_command


def test_purge_purgatory(settings):
    settings.UMAP_PURGATORY_ROOT = tempfile.mkdtemp()
    root = Path(settings.UMAP_PURGATORY_ROOT)
    old = root / "old.json"
    old.write_text("{}")
    stat = old.stat()
    os.utime(old, times=(stat.st_mtime - 31 * 86400, stat.st_mtime - 31 * 86400))
    recent = root / "recent.json"
    recent.write_text("{}")
    stat = recent.stat()
    os.utime(recent, times=(stat.st_mtime - 8 * 86400, stat.st_mtime - 8 * 86400))
    now = root / "now.json"
    now.write_text("{}")
    assert {f.name for f in root.iterdir()} == {"old.json", "recent.json", "now.json"}
    call_command("purge_purgatory")
    assert {f.name for f in root.iterdir()} == {"recent.json", "now.json"}
    call_command("purge_purgatory", "--days=7")
    assert {f.name for f in root.iterdir()} == {"now.json"}



import json
import re
from pathlib import Path
from time import sleep

from playwright.sync_api import expect

from umap.models import DataLayer

from ..base import DataLayerFactory, MapFactory

DATALAYER_UPDATE = re.compile(r".*/datalayer/update/.*")


def test_created_markers_are_merged(context, live_server, tilelayer):
    # Let's create a new map with an empty datalayer
    map = MapFactory(name="server-side merge")
    datalayer = DataLayerFactory(map=map, edit_status=DataLayer.ANONYMOUS, data={})

    # Now navigate to this map and create marker
    page_one = context.new_page()
    page_one.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    save_p1 = page_one.get_by_role("button", name="Save")
    expect(save_p1).to_be_visible()

    # Click on the Draw a marker button on a new map.
    create_marker_p1 = page_one.get_by_title("Draw a marker")
    expect(create_marker_p1).to_be_visible()
    create_marker_p1.click()

    # Check no marker is present by default.
    marker_pane_p1 = page_one.locator(".leaflet-marker-pane > div")
    expect(marker_pane_p1).to_have_count(0)

    # Click on the map, it will place a marker at the given position.
    map_el_p1 = page_one.locator("#map")
    map_el_p1.click(position={"x": 200, "y": 200})
    expect(marker_pane_p1).to_have_count(1)

    with page_one.expect_response(DATALAYER_UPDATE):
        save_p1.click()
        # Prevent two layers to be saved on the same second, as we compare them based
        # on time in case of conflict. FIXME do not use time for comparison.
        sleep(1)
    assert DataLayer.objects.get(pk=datalayer.pk).settings == {
        "browsable": True,
        "displayOnLoad": True,
        "name": "test datalayer",
        "editMode": "advanced",
        "inCaption": True,
    }

    # Now navigate to this map from another tab
    page_two = context.new_page()

    page_two.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    save_p2 = page_two.get_by_role("button", name="Save")
    expect(save_p2).to_be_visible()

    # Click on the Draw a marker button on a new map.
    create_marker_p2 = page_two.get_by_title("Draw a marker")
    expect(create_marker_p2).to_be_visible()
    create_marker_p2.click()

    # Check that the marker created in the orther tab is present.
    marker_pane_p2 = page_two.locator(".leaflet-marker-pane > div")
    expect(marker_pane_p2).to_have_count(1)

    # Click on the map, it will place a marker at the given position.
    map_el_p2 = page_two.locator("#map")
    map_el_p2.click(position={"x": 220, "y": 220})
    expect(marker_pane_p2).to_have_count(2)

    with page_two.expect_response(DATALAYER_UPDATE):
        save_p2.click()
        sleep(1)
    # No change after the save
    expect(marker_pane_p2).to_have_count(2)
    assert DataLayer.objects.get(pk=datalayer.pk).settings == {
        "browsable": True,
        "displayOnLoad": True,
        "name": "test datalayer",
        "inCaption": True,
        "editMode": "advanced",
    }

    # Now create another marker in the first tab
    create_marker_p1.click()
    map_el_p1.click(position={"x": 150, "y": 150})
    expect(marker_pane_p1).to_have_count(2)
    with page_one.expect_response(DATALAYER_UPDATE):
        save_p1.click()
    # Should now get the other marker too
    expect(marker_pane_p1).to_have_count(3)
    assert DataLayer.objects.get(pk=datalayer.pk).settings == {
        "browsable": True,
        "displayOnLoad": True,
        "name": "test datalayer",
        "inCaption": True,
        "editMode": "advanced",
        "id": str(datalayer.pk),
        "permissions": {"edit_status": 1},
    }

    # And again
    create_marker_p1.click()
    map_el_p1.click(position={"x": 180, "y": 150})
    expect(marker_pane_p1).to_have_count(4)
    with page_one.expect_response(DATALAYER_UPDATE):
        save_p1.click()
        sleep(1)
    # Should now get the other marker too
    assert DataLayer.objects.get(pk=datalayer.pk).settings == {
        "browsable": True,
        "displayOnLoad": True,
        "name": "test datalayer",
        "inCaption": True,
        "editMode": "advanced",
        "id": str(datalayer.pk),
        "permissions": {"edit_status": 1},
    }
    expect(marker_pane_p1).to_have_count(4)

    # And again from the second tab
    expect(marker_pane_p2).to_have_count(2)
    create_marker_p2.click()
    map_el_p2.click(position={"x": 250, "y": 150})
    expect(marker_pane_p2).to_have_count(3)
    with page_two.expect_response(DATALAYER_UPDATE):
        save_p2.click()
        sleep(1)
    # Should now get the other markers too
    assert DataLayer.objects.get(pk=datalayer.pk).settings == {
        "browsable": True,
        "displayOnLoad": True,
        "name": "test datalayer",
        "inCaption": True,
        "editMode": "advanced",
        "id": str(datalayer.pk),
        "permissions": {"edit_status": 1},
    }
    expect(marker_pane_p2).to_have_count(5)


def test_empty_datalayers_can_be_merged(context, live_server, tilelayer):
    # Let's create a new map with an empty datalayer
    map = MapFactory(name="server-side merge")
    DataLayerFactory(map=map, edit_status=DataLayer.ANONYMOUS, data={})

    # Open two tabs at the same time, on the same empty map
    page_one = context.new_page()
    page_one.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    page_two = context.new_page()
    page_two.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    save_p1 = page_one.get_by_role("button", name="Save")
    expect(save_p1).to_be_visible()

    # Click on the Draw a marker button on a new map.
    create_marker_p1 = page_one.get_by_title("Draw a marker")
    expect(create_marker_p1).to_be_visible()
    create_marker_p1.click()

    # Check no marker is present by default.
    marker_pane_p1 = page_one.locator(".leaflet-marker-pane > div")
    expect(marker_pane_p1).to_have_count(0)

    # Click on the map, it will place a marker at the given position.
    map_el_p1 = page_one.locator("#map")
    map_el_p1.click(position={"x": 200, "y": 200})
    expect(marker_pane_p1).to_have_count(1)

    with page_one.expect_response(DATALAYER_UPDATE):
        save_p1.click()
        sleep(1)

    save_p2 = page_two.get_by_role("button", name="Save")
    expect(save_p2).to_be_visible()

    # Click on the Draw a marker button on a new map.
    create_marker_p2 = page_two.get_by_title("Draw a marker")
    expect(create_marker_p2).to_be_visible()
    create_marker_p2.click()

    marker_pane_p2 = page_two.locator(".leaflet-marker-pane > div")

    # Click on the map, it will place a marker at the given position.
    map_el_p2 = page_two.locator("#map")
    map_el_p2.click(position={"x": 220, "y": 220})
    expect(marker_pane_p2).to_have_count(1)

    # Save p1 and p2 at the same time
    with page_two.expect_response(DATALAYER_UPDATE):
        save_p2.click()
        sleep(1)

    expect(marker_pane_p2).to_have_count(2)


def test_same_second_edit_doesnt_conflict(context, live_server, tilelayer):
    # Let's create a new map with an empty datalayer
    map = MapFactory(name="server-side merge")
    datalayer = DataLayerFactory(map=map, edit_status=DataLayer.ANONYMOUS, data={})

    # Open the created map on two pages.
    page_one = context.new_page()
    page_one.goto(f"{live_server.url}{map.get_absolute_url()}?edit")
    page_two = context.new_page()
    page_two.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    save_p1 = page_one.get_by_role("button", name="Save")
    expect(save_p1).to_be_visible()

    save_p2 = page_two.get_by_role("button", name="Save")
    expect(save_p2).to_be_visible()

    # Create a point on the first map
    create_marker_p1 = page_one.get_by_title("Draw a marker")
    expect(create_marker_p1).to_be_visible()
    create_marker_p1.click()

    # Check no marker is present by default.
    marker_pane_p1 = page_one.locator(".leaflet-marker-pane > div")
    expect(marker_pane_p1).to_have_count(0)

    # Click on the map, it will place a marker at the given position.
    map_el_p1 = page_one.locator("#map")
    map_el_p1.click(position={"x": 200, "y": 200})
    expect(marker_pane_p1).to_have_count(1)

    # And add one on the second map as well.
    create_marker_p2 = page_two.get_by_title("Draw a marker")
    expect(create_marker_p2).to_be_visible()
    create_marker_p2.click()

    marker_pane_p2 = page_two.locator(".leaflet-marker-pane > div")

    # Click on the map, it will place a marker at the given position.
    map_el_p2 = page_two.locator("#map")
    map_el_p2.click(position={"x": 220, "y": 220})
    expect(marker_pane_p2).to_have_count(1)

    # Save the two tabs at the same time
    with page_one.expect_response(DATALAYER_UPDATE):
        save_p1.click()
        sleep(0.2)  # Needed to avoid having multiple requests coming at the same time.
        save_p2.click()

    # Now create another marker in the first tab
    create_marker_p1.click()
    map_el_p1.click(position={"x": 150, "y": 150})
    expect(marker_pane_p1).to_have_count(2)
    with page_one.expect_response(DATALAYER_UPDATE):
        save_p1.click()

    # Should now get the other marker too
    expect(marker_pane_p1).to_have_count(3)
    assert DataLayer.objects.get(pk=datalayer.pk).settings == {
        "browsable": True,
        "displayOnLoad": True,
        "name": "test datalayer",
        "inCaption": True,
        "editMode": "advanced",
        "id": str(datalayer.pk),
        "permissions": {"edit_status": 1},
    }


def test_should_display_alert_on_conflict(context, live_server, datalayer, openmap):
    # Open the map on two pages.
    page_one = context.new_page()
    page_one.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit")
    page_two = context.new_page()
    page_two.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit")

    page_one.locator(".leaflet-marker-icon").click(modifiers=["Shift"])
    page_one.locator('input[name="name"]').fill("new name")
    with page_one.expect_response(re.compile(r".*/datalayer/update/.*")):
        page_one.get_by_role("button", name="Save").click()

    page_two.locator(".leaflet-marker-icon").click(modifiers=["Shift"])
    page_two.locator('input[name="name"]').fill("custom name")
    with page_two.expect_response(re.compile(r".*/datalayer/update/.*")):
        page_two.get_by_role("button", name="Save").click()
    saved = DataLayer.objects.last()
    data = json.loads(Path(saved.geojson.path).read_text())
    assert data["features"][0]["properties"]["name"] == "new name"
    expect(page_two.get_by_text("Whoops! Other contributor(s) changed")).to_be_visible()
    with page_two.expect_response(re.compile(r".*/datalayer/update/.*")):
        page_two.get_by_text("Keep your changes and loose theirs").click()
    saved = DataLayer.objects.last()
    data = json.loads(Path(saved.geojson.path).read_text())
    assert data["features"][0]["properties"]["name"] == "custom name"



#!/usr/bin/env python

import asyncio
import logging
import uuid
from collections import defaultdict
from typing import Literal, Optional, Union

import websockets
from django.conf import settings
from django.core.signing import TimestampSigner
from pydantic import BaseModel, Field, RootModel, ValidationError
from websockets import WebSocketClientProtocol
from websockets.server import serve


class Connections:
    def __init__(self) -> None:
        self._connections: set[WebSocketClientProtocol] = set()
        self._ids: dict[WebSocketClientProtocol, str] = dict()

    def join(self, websocket: WebSocketClientProtocol) -> str:
        self._connections.add(websocket)
        _id = str(uuid.uuid4())
        self._ids[websocket] = _id
        return _id

    def leave(self, websocket: WebSocketClientProtocol) -> None:
        self._connections.remove(websocket)
        del self._ids[websocket]

    def get(self, id) -> WebSocketClientProtocol:
        # use an iterator to stop iterating as soon as we found
        return next(k for k, v in self._ids.items() if v == id)

    def get_id(self, websocket: WebSocketClientProtocol):
        return self._ids[websocket]

    def get_other_peers(
        self, websocket: WebSocketClientProtocol
    ) -> set[WebSocketClientProtocol]:
        return self._connections - {websocket}

    def get_all_peers(self) -> set[WebSocketClientProtocol]:
        return self._connections


# Contains the list of websocket connections handled by this process.
# It's a mapping of map_id to a set of the active websocket connections
CONNECTIONS: defaultdict[int, Connections] = defaultdict(Connections)


class JoinRequest(BaseModel):
    kind: Literal["JoinRequest"] = "JoinRequest"
    token: str


class OperationMessage(BaseModel):
    """Message sent from one peer to all the others"""

    kind: Literal["OperationMessage"] = "OperationMessage"
    verb: Literal["upsert", "update", "delete"]
    subject: Literal["map", "datalayer", "feature"]
    metadata: Optional[dict] = None
    key: Optional[str] = None


class PeerMessage(BaseModel):
    """Message sent from a specific peer to another one"""

    kind: Literal["PeerMessage"] = "PeerMessage"
    sender: str
    recipient: str
    # The message can be whatever the peers want. It's not checked by the server.
    message: dict


class ServerRequest(BaseModel):
    """A request towards the server"""

    kind: Literal["Server"] = "Server"
    action: Literal["list-peers"]


class Request(RootModel):
    """Any message coming from the websocket should be one of these, and will be rejected otherwise."""

    root: Union[ServerRequest, PeerMessage, OperationMessage] = Field(
        discriminator="kind"
    )


class JoinResponse(BaseModel):
    """Server response containing the list of peers"""

    kind: Literal["JoinResponse"] = "JoinResponse"
    peers: list
    uuid: str


class ListPeersResponse(BaseModel):
    kind: Literal["ListPeersResponse"] = "ListPeersResponse"
    peers: list


async def join_and_listen(
    map_id: int, permissions: list, user: str | int, websocket: WebSocketClientProtocol
):
    """Join a "room" with other connected peers, and wait for messages."""
    logging.debug(f"{user} joined room #{map_id}")
    connections: Connections = CONNECTIONS[map_id]
    _id: str = connections.join(websocket)

    # Assign an ID to the joining peer and return it the list of connected peers.
    peers: list[WebSocketClientProtocol] = [
        connections.get_id(p) for p in connections.get_all_peers()
    ]
    response = JoinResponse(uuid=_id, peers=peers)
    await websocket.send(response.model_dump_json())

    # Notify all other peers of the new list of connected peers.
    message = ListPeersResponse(peers=peers)
    websockets.broadcast(
        connections.get_other_peers(websocket), message.model_dump_json()
    )

    try:
        async for raw_message in websocket:
            # recompute the peers list at the time of message-sending.
            # as doing so beforehand would miss new connections
            other_peers = connections.get_other_peers(websocket)
            try:
                incoming = Request.model_validate_json(raw_message)
            except ValidationError as e:
                error = f"An error occurred when receiving the following message: {raw_message!r}"
                logging.error(error, e)
            else:
                match incoming.root:
                    # Broadcast all operation messages to connected peers
                    case OperationMessage():
                        websockets.broadcast(other_peers, raw_message)

                    # Send peer messages to the proper peer
                    case PeerMessage(recipient=_id):
                        peer = connections.get(_id)
                        if peer:
                            await peer.send(raw_message)

    finally:
        # On disconnect, remove the connection from the pool
        connections.leave(websocket)

        # TODO: refactor this in a separate method.
        # Notify all other peers of the new list of connected peers.
        peers = [connections.get_id(p) for p in connections.get_all_peers()]
        message = ListPeersResponse(peers=peers)
        websockets.broadcast(
            connections.get_other_peers(websocket), message.model_dump_json()
        )


async def handler(websocket: WebSocketClientProtocol):
    """Main WebSocket handler.

    Check if the permission is granted and let the peer enter a room.
    """
    raw_message = await websocket.recv()

    # The first event should always be 'join'
    message: JoinRequest = JoinRequest.model_validate_json(raw_message)
    signed = TimestampSigner().unsign_object(message.token, max_age=30)
    user, map_id, permissions = signed.values()

    # Check if permissions for this map have been granted by the server
    if "edit" in signed["permissions"]:
        await join_and_listen(map_id, permissions, user, websocket)


def run(host: str, port: int):
    if not settings.WEBSOCKET_ENABLED:
        msg = (
            "WEBSOCKET_ENABLED should be set to True to run the WebSocket Server. "
            "See the documentation at "
            "https://docs.umap-project.org/en/stable/config/settings/#websocket_enabled "
            "for more information."
        )
        print(msg)
        exit(1)

    async def _serve():
        async with serve(handler, host, port):
            logging.debug(f"Waiting for connections on {host}:{port}")
            await asyncio.Future()  # run forever

    asyncio.run(_serve())



# Generated by Django 2.1.5 on 2019-04-07 07:19

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0005_remove_map_tilelayer"),
    ]

    operations = [
        migrations.AlterField(
            model_name="map",
            name="settings",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True, default=dict, null=True, verbose_name="settings"
            ),
        ),
    ]



import json
from pathlib import Path

import pytest
from playwright.sync_api import expect

from ..base import DataLayerFactory

pytestmark = pytest.mark.django_db


def test_basic_circles_layer(map, live_server, page):
    path = Path(__file__).parent.parent / "fixtures/test_circles_layer.geojson"
    data = json.loads(path.read_text())
    DataLayerFactory(data=data, map=map)
    page.goto(f"{live_server.url}{map.get_absolute_url()}#12/47.2210/-1.5621")
    paths = page.locator("path")
    expect(paths).to_have_count(10)
    # Last arc curve command
    assert (
        page.locator("[data-feature=c1160]")
        .get_attribute("d")
        .endswith("a40,40 0 1,0 -80,0 ")
    )
    assert (
        page.locator("[data-feature=ca676]")
        .get_attribute("d")
        .endswith("a31,31 0 1,0 -62,0 ")
    )
    assert (
        page.locator("[data-feature=cap64]")
        .get_attribute("d")
        .endswith("a10,10 0 1,0 -20,0 ")
    )
    assert (
        page.locator("[data-feature=cap27]")
        .get_attribute("d")
        .endswith("a6,6 0 1,0 -12,0 ")
    )
    assert (
        page.locator("[data-feature=capa8]")
        .get_attribute("d")
        .endswith("a4,4 0 1,0 -8,0 ")
    )
    assert (
        page.locator("[data-feature=capa6]")
        .get_attribute("d")
        .endswith("a3,3 0 1,0 -6,0 ")
    )
    assert (
        page.locator("[data-feature=capa4]")
        .get_attribute("d")
        .endswith("a3,3 0 1,0 -6,0 ")
    )
    assert (
        page.locator("[data-feature=capa3]")
        .get_attribute("d")
        .endswith("a2,2 0 1,0 -4,0 ")
    )
    assert (
        page.locator("[data-feature=capa2]")
        .get_attribute("d")
        .endswith("a2,2 0 1,0 -4,0 ")
    )
    assert (
        page.locator("[data-feature=capa0]")
        .get_attribute("d")
        .endswith("a2,2 0 1,0 -4,0 ")
    )


def test_can_draw_new_circles(openmap, live_server, page):
    path = Path(__file__).parent.parent / "fixtures/test_circles_layer.geojson"
    data = json.loads(path.read_text())
    DataLayerFactory(data=data, map=openmap)
    page.goto(f"{live_server.url}{openmap.get_absolute_url()}?edit#12/47.2210/-1.5621")
    paths = page.locator("path")
    expect(paths).to_have_count(10)
    page.get_by_title("Draw a marker").click()
    page.locator("#map").click(position={"x": 200, "y": 200})
    expect(paths).to_have_count(11)



import json
from pathlib import Path

import pytest
from playwright.sync_api import expect

from ..base import DataLayerFactory

pytestmark = pytest.mark.django_db


def test_basic_choropleth_map_with_default_color(map, live_server, page):
    path = Path(__file__).parent.parent / "fixtures/choropleth_region_chomage.geojson"
    data = json.loads(path.read_text())
    DataLayerFactory(data=data, map=map)
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    # Hauts-de-France
    expect(page.locator("path[fill='#08519c']")).to_have_count(1)
    # Occitanie
    expect(page.locator("path[fill='#3182bd']")).to_have_count(1)
    # Grand-Est, PACA
    expect(page.locator("path[fill='#6baed6']")).to_have_count(2)
    # Bourgogne-Franche-Comté, Centre-Val-de-Loire, IdF, Normandie, Corse, Nouvelle-Aquitaine
    expect(page.locator("path[fill='#bdd7e7']")).to_have_count(6)
    # Bretagne, Pays de la Loire, AURA
    expect(page.locator("path[fill='#eff3ff']")).to_have_count(3)


def test_basic_choropleth_map_with_custom_brewer(openmap, live_server, page):
    path = Path(__file__).parent.parent / "fixtures/choropleth_region_chomage.geojson"
    data = json.loads(path.read_text())

    # Change brewer at load
    data["_umap_options"]["choropleth"]["brewer"] = "Reds"
    DataLayerFactory(data=data, map=openmap)

    page.goto(f"{live_server.url}{openmap.get_absolute_url()}")
    # Hauts-de-France
    expect(page.locator("path[fill='#a50f15']")).to_have_count(1)
    # Occitanie
    expect(page.locator("path[fill='#de2d26']")).to_have_count(1)
    # Grand-Est, PACA
    expect(page.locator("path[fill='#fb6a4a']")).to_have_count(2)
    # Bourgogne-Franche-Comté, Centre-Val-de-Loire, IdF, Normandie, Corse, Nouvelle-Aquitaine
    expect(page.locator("path[fill='#fcae91']")).to_have_count(6)
    # Bretagne, Pays de la Loire, AURA
    expect(page.locator("path[fill='#fee5d9']")).to_have_count(3)

    # Now change brewer from UI
    page.get_by_role("button", name="Edit").click()
    page.get_by_role("link", name="Manage layers").click()
    page.locator(".panel").get_by_title("Edit", exact=True).click()
    page.get_by_text("Choropleth: settings").click()
    page.locator('select[name="brewer"]').select_option("Greens")

    # Hauts-de-France
    expect(page.locator("path[fill='#006d2c']")).to_have_count(1)
    # Occitanie
    expect(page.locator("path[fill='#31a354']")).to_have_count(1)
    # Grand-Est, PACA
    expect(page.locator("path[fill='#74c476']")).to_have_count(2)
    # Bourgogne-Franche-Comté, Centre-Val-de-Loire, IdF, Normandie, Corse, Nouvelle-Aquitaine
    expect(page.locator("path[fill='#bae4b3']")).to_have_count(6)
    # Bretagne, Pays de la Loire, AURA
    expect(page.locator("path[fill='#edf8e9']")).to_have_count(3)


def test_basic_choropleth_map_with_custom_classes(openmap, live_server, page):
    path = Path(__file__).parent.parent / "fixtures/choropleth_region_chomage.geojson"
    data = json.loads(path.read_text())

    # Change brewer at load
    data["_umap_options"]["choropleth"]["classes"] = 6
    DataLayerFactory(data=data, map=openmap)

    page.goto(f"{live_server.url}{openmap.get_absolute_url()}")

    # Hauts-de-France
    expect(page.locator("path[fill='#08519c']")).to_have_count(1)
    # Occitanie
    expect(page.locator("path[fill='#3182bd']")).to_have_count(1)
    # PACA
    expect(page.locator("path[fill='#6baed6']")).to_have_count(1)
    # Grand-Est
    expect(page.locator("path[fill='#9ecae1']")).to_have_count(1)
    # Bourgogne-Franche-Comté, Centre-Val-de-Loire, IdF, Normandie, Corse, Nouvelle-Aquitaine
    expect(page.locator("path[fill='#c6dbef']")).to_have_count(6)
    # Bretagne, Pays de la Loire, AURA
    expect(page.locator("path[fill='#eff3ff']")).to_have_count(3)



import re

import pytest
from playwright.sync_api import expect

from umap.models import Star

pytestmark = pytest.mark.django_db


def test_star_control_is_visible_if_logged_in(map, live_server, page, login, user):
    login(user)
    assert not Star.objects.count()
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    page.get_by_title("More controls").click()
    control = page.locator(".leaflet-control-star")
    expect(control).to_be_visible()
    with page.expect_response(re.compile(".*/star/")):
        control.click()
    assert Star.objects.count() == 1


def test_no_star_control_if_not_logged_in(map, live_server, page):
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    page.get_by_title("More controls").click()
    control = page.locator(".leaflet-control-star")
    expect(control).to_be_hidden()



import json
import shutil
import tempfile
from copy import deepcopy
from pathlib import Path

import pytest
from django.core.management import call_command


@pytest.fixture
def staticfiles(settings):
    static_root = tempfile.mkdtemp(prefix="test_static")
    settings.STATIC_ROOT = static_root
    # Make sure settings are properly reset after the test
    settings.STORAGES = deepcopy(settings.STORAGES)
    settings.STORAGES["staticfiles"]["BACKEND"] = (
        "umap.storage.UmapManifestStaticFilesStorage"
    )
    try:
        call_command("collectstatic", "--noinput")
        yield
    finally:
        shutil.rmtree(static_root)


def test_collectstatic_ran_successfully_with_hashes(settings, staticfiles):
    static_root = settings.STATIC_ROOT
    manifest = Path(static_root) / "staticfiles.json"
    assert manifest.exists()
    json_manifest = json.loads(manifest.read_text())
    assert "hash" in json_manifest.keys()
    assert "umap/base.css" in json_manifest["paths"]
    # Hash + the dot ("umap/base.<hash>.css").
    md5_hash_lenght = 12 + 1
    # The value of the manifest must contain the hash (length).
    assert (
        len(json_manifest["paths"]["umap/base.css"])
        == len("umap/base.css") + md5_hash_lenght
    )



import re

import pytest
from playwright.sync_api import expect

from umap.models import Map

from ..base import DataLayerFactory, MapFactory

DATALAYER_UPDATE = re.compile(r".*/datalayer/update/.*")


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_markers(
    new_page, live_server, websocket_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    a_marker_pane = peerA.locator(".leaflet-marker-pane > div")
    b_marker_pane = peerB.locator(".leaflet-marker-pane > div")
    expect(a_marker_pane).to_have_count(0)
    expect(b_marker_pane).to_have_count(0)

    # Add a marker from peer A
    a_create_marker = peerA.get_by_title("Draw a marker")
    expect(a_create_marker).to_be_visible()
    a_create_marker.click()

    a_map_el = peerA.locator("#map")
    a_map_el.click(position={"x": 220, "y": 220})
    expect(a_marker_pane).to_have_count(1)
    expect(b_marker_pane).to_have_count(1)
    peerA.locator("body").type("Synced name")
    peerA.locator("body").press("Escape")

    peerB.locator(".leaflet-marker-icon").first.click()
    peerB.get_by_role("link", name="Toggle edit mode (⇧+Click)").click()
    expect(peerB.locator('input[name="name"]')).to_have_value("Synced name")

    a_first_marker = peerA.locator("div:nth-child(4) > div:nth-child(2)").first
    b_first_marker = peerB.locator("div:nth-child(4) > div:nth-child(2)").first

    # Add a second marker from peer B
    b_create_marker = peerB.get_by_title("Draw a marker")
    expect(b_create_marker).to_be_visible()
    b_create_marker.click()

    b_map_el = peerB.locator("#map")
    b_map_el.click(position={"x": 225, "y": 225})
    expect(a_marker_pane).to_have_count(2)
    expect(b_marker_pane).to_have_count(2)

    # Drag a marker on peer B and check that it moved on peer A
    assert a_first_marker.bounding_box() == b_first_marker.bounding_box()
    b_old_bbox = b_first_marker.bounding_box()
    b_first_marker.drag_to(b_map_el, target_position={"x": 250, "y": 250})

    assert b_old_bbox is not b_first_marker.bounding_box()
    assert a_first_marker.bounding_box() == b_first_marker.bounding_box()

    # Delete a marker from peer A and check it's been deleted on peer B
    a_first_marker.click(button="right")
    peerA.get_by_role("button", name="Delete this feature").click()
    peerA.locator("dialog").get_by_role("button", name="OK").click()
    expect(a_marker_pane).to_have_count(1)
    expect(b_marker_pane).to_have_count(1)


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_polygons(
    context, live_server, websocket_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = context.new_page()
    peerA.goto(f"{live_server.url}{map.get_absolute_url()}?edit")
    peerB = context.new_page()
    peerB.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    b_map_el = peerB.locator("#map")

    # Click on the Draw a polygon button on a new map.
    create_line = peerA.locator(".leaflet-control-toolbar ").get_by_title(
        "Draw a polygon"
    )
    create_line.click()

    a_polygons = peerA.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    b_polygons = peerB.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    expect(a_polygons).to_have_count(0)
    expect(b_polygons).to_have_count(0)

    # Click on the map, it will create a polygon.
    map = peerA.locator("#map")
    map.click(position={"x": 200, "y": 200})
    map.click(position={"x": 100, "y": 200})
    map.click(position={"x": 100, "y": 100})
    map.click(position={"x": 100, "y": 100})

    # It is created on peerA, but not yet synced
    expect(a_polygons).to_have_count(1)
    expect(b_polygons).to_have_count(0)

    # Escaping the edition syncs
    peerA.keyboard.press("Escape")
    expect(a_polygons).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # change the geometry by moving a point on peer B
    a_polygon = peerA.locator("path")
    b_polygon = peerB.locator("path")
    b_polygon_bbox_t1 = b_polygon.bounding_box()
    a_polygon_bbox_t1 = a_polygon.bounding_box()
    assert b_polygon_bbox_t1 == a_polygon_bbox_t1

    b_polygon.click()
    peerB.get_by_role("link", name="Toggle edit mode (⇧+Click)").click()

    edited_vertex = peerB.locator("div:nth-child(6)").first
    edited_vertex.drag_to(b_map_el, target_position={"x": 233, "y": 126})
    peerB.keyboard.press("Escape")

    b_polygon_bbox_t2 = b_polygon.bounding_box()
    a_polygon_bbox_t2 = a_polygon.bounding_box()

    assert b_polygon_bbox_t2 != b_polygon_bbox_t1
    assert b_polygon_bbox_t2 == a_polygon_bbox_t2

    # Move the polygon on peer B and check it moved also on peer A
    b_polygon.click()
    peerB.get_by_role("link", name="Toggle edit mode (⇧+Click)").click()

    b_polygon.drag_to(b_map_el, target_position={"x": 400, "y": 400})
    peerB.keyboard.press("Escape")
    b_polygon_bbox_t3 = b_polygon.bounding_box()
    a_polygon_bbox_t3 = a_polygon.bounding_box()

    assert b_polygon_bbox_t3 != b_polygon_bbox_t2
    assert b_polygon_bbox_t3 == a_polygon_bbox_t3

    # Delete a polygon from peer A and check it's been deleted on peer B
    a_polygon.click(button="right")
    peerA.get_by_role("button", name="Delete this feature").click()
    peerA.locator("dialog").get_by_role("button", name="OK").click()
    expect(a_polygons).to_have_count(0)
    expect(b_polygons).to_have_count(0)


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_map_properties(
    context, live_server, websocket_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = context.new_page()
    peerA.goto(f"{live_server.url}{map.get_absolute_url()}?edit")
    peerB = context.new_page()
    peerB.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    # Name change is synced
    peerA.get_by_role("link", name="Edit map name and caption").click()
    peerA.locator('input[name="name"]').click()
    peerA.locator('input[name="name"]').fill("it syncs!")

    expect(peerB.locator(".map-name").last).to_have_text("it syncs!")

    # Zoom control is synced
    peerB.get_by_role("link", name="Map advanced properties").click()
    peerB.locator("summary").filter(has_text="User interface options").click()
    peerB.locator("div").filter(
        has_text=re.compile(r"^Display the zoom control")
    ).locator("label").nth(2).click()

    expect(peerA.locator(".leaflet-control-zoom")).to_be_hidden()


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_datalayer_properties(
    context, live_server, websocket_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = context.new_page()
    peerA.goto(f"{live_server.url}{map.get_absolute_url()}?edit")
    peerB = context.new_page()
    peerB.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    # Layer addition, name and type are synced
    peerA.get_by_role("link", name="Manage layers").click()
    peerA.get_by_role("button", name="Add a layer").click()
    peerA.locator('input[name="name"]').click()
    peerA.locator('input[name="name"]').fill("synced layer!")
    peerA.get_by_role("combobox").select_option("Choropleth")
    peerA.locator("body").press("Escape")

    peerB.get_by_role("link", name="Manage layers").click()
    peerB.get_by_role("button", name="Edit").first.click()
    expect(peerB.locator('input[name="name"]')).to_have_value("synced layer!")
    expect(peerB.get_by_role("combobox")).to_have_value("Choropleth")


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_cloned_polygons(
    context, live_server, websocket_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = context.new_page()
    peerA.goto(f"{live_server.url}{map.get_absolute_url()}?edit")
    peerB = context.new_page()
    peerB.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    b_map_el = peerB.locator("#map")

    # Click on the Draw a polygon button on a new map.
    create_line = peerA.locator(".leaflet-control-toolbar ").get_by_title(
        "Draw a polygon"
    )
    create_line.click()

    a_polygons = peerA.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    b_polygons = peerB.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    expect(a_polygons).to_have_count(0)
    expect(b_polygons).to_have_count(0)

    # Click on the map, it will create a polygon.
    map = peerA.locator("#map")
    map.click(position={"x": 200, "y": 200})
    map.click(position={"x": 100, "y": 200})
    map.click(position={"x": 100, "y": 100})
    map.click(position={"x": 200, "y": 100})
    map.click(position={"x": 200, "y": 100})

    # Escaping the edition syncs
    peerA.keyboard.press("Escape")
    expect(a_polygons).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # Save from peer A
    peerA.get_by_role("button", name="Save").click()

    b_polygon = peerB.locator("path")

    # Clone on peer B and save
    b_polygon.click(button="right")
    peerB.get_by_role("button", name="Clone this feature").click()

    expect(peerB.locator("path")).to_have_count(2)

    peerB.locator("path").nth(1).drag_to(b_map_el, target_position={"x": 400, "y": 400})
    peerB.locator("path").nth(1).click()
    peerB.locator("summary").filter(has_text="Shape properties").click()
    peerB.locator(".header > a:nth-child(2)").first.click()
    peerB.get_by_title("Orchid", exact=True).first.click()
    peerB.locator("#map").press("Escape")
    peerB.get_by_role("button", name="Save").click()

    expect(peerB.locator("path")).to_have_count(2)


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_late_joining_peer(
    new_page, live_server, websocket_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create first peer (A) and have it join immediately
    peerA = new_page("Page A")
    peerA.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    # Add a marker from peer A
    a_create_marker = peerA.get_by_title("Draw a marker")
    expect(a_create_marker).to_be_visible()
    a_create_marker.click()

    a_map_el = peerA.locator("#map")
    a_map_el.click(position={"x": 220, "y": 220})
    peerA.locator("body").type("First marker")
    peerA.locator("body").press("Escape")

    # Add a polygon from peer A
    create_polygon = peerA.locator(".leaflet-control-toolbar ").get_by_title(
        "Draw a polygon"
    )
    create_polygon.click()

    a_map_el.click(position={"x": 200, "y": 200})
    a_map_el.click(position={"x": 100, "y": 200})
    a_map_el.click(position={"x": 100, "y": 100})
    a_map_el.click(position={"x": 200, "y": 100})
    a_map_el.click(position={"x": 200, "y": 100})
    peerA.keyboard.press("Escape")

    # Now create peer B and have it join
    peerB = new_page("Page B")
    peerB.goto(f"{live_server.url}{map.get_absolute_url()}?edit")

    # Check if peer B has received all the updates
    b_marker_pane = peerB.locator(".leaflet-marker-pane > div")
    b_polygons = peerB.locator(".leaflet-overlay-pane path[fill='DarkBlue']")

    expect(b_marker_pane).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # Verify marker properties
    peerB.locator(".leaflet-marker-icon").first.click()
    peerB.get_by_role("link", name="Toggle edit mode (⇧+Click)").click()
    expect(peerB.locator('input[name="name"]')).to_have_value("First marker")

    # Verify polygon exists (we've already checked the count)
    b_polygon = peerB.locator("path")
    expect(b_polygon).to_be_visible()

    # Optional: Verify polygon properties if you have any specific ones set

    # Clean up: close edit mode
    peerB.locator("body").press("Escape")



from umap.websocket_server import OperationMessage, PeerMessage, Request, ServerRequest


def test_messages_are_parsed_correctly():
    server = Request.model_validate(dict(kind="Server", action="list-peers")).root
    assert type(server) is ServerRequest

    operation = Request.model_validate(
        dict(
            kind="OperationMessage",
            verb="upsert",
            subject="map",
            metadata={},
            key="key",
        )
    ).root
    assert type(operation) is OperationMessage

    peer_message = Request.model_validate(
        dict(kind="PeerMessage", sender="Alice", recipient="Bob", message={})
    ).root
    assert type(peer_message) is PeerMessage



import pytest
from playwright.sync_api import expect

from ..base import DataLayerFactory

pytestmark = pytest.mark.django_db


DATALAYER_DATA = {
    "type": "FeatureCollection",
    "_umap_options": {
        "name": "Calque 1",
        "type": "Cluster",
        "cluster": {},
        "browsable": True,
        "inCaption": True,
        "displayOnLoad": True,
    },
    "features": [
        {
            "type": "Feature",
            "properties": {"name": "one point in france"},
            "geometry": {"type": "Point", "coordinates": [3.339844, 46.920255]},
        },
        {
            "type": "Feature",
            "properties": {
                "name": "one another point in france in same position",
                "description": "can you see me ?",
            },
            "geometry": {"type": "Point", "coordinates": [3.339844, 46.920255]},
        },
        {
            "type": "Feature",
            "properties": {
                "name": "again one another point",
                "description": "and me ?",
            },
            "geometry": {"type": "Point", "coordinates": [3.34, 46.1]},
        },
    ],
}


def test_can_open_feature_on_browser_click(live_server, page, map):
    map.settings["properties"]["onLoadPanel"] = "databrowser"
    map.save()
    DataLayerFactory(map=map, data=DATALAYER_DATA)
    page.goto(f"{live_server.url}{map.get_absolute_url()}#7/46.920/3.340")
    page.get_by_text("one another point in france in same position").click()
    expect(page.get_by_text("can you see me ?")).to_be_visible()
    page.get_by_text("again one another point").click()
    expect(page.get_by_text("and me ?")).to_be_visible()



import pytest
from django.core.management import call_command

from umap.models import Map

pytestmark = pytest.mark.django_db


def test_can_delete_tilelayer_from_settings(map):
    map.settings["properties"]["tilelayer"] = {
        "name": "My TileLayer",
        "maxZoom": 18,
        "minZoom": 0,
        "attribution": "My attribution",
        "url_template": "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
    }
    map.save()
    # Make sure its saved
    map = Map.objects.get(pk=map.pk)
    assert "tilelayer" in map.settings["properties"]
    call_command(
        "clean_tilelayer", "http://{s}.foo.bar.baz/{z}/{x}/{y}.png", "--no-input"
    )
    map = Map.objects.get(pk=map.pk)
    assert "tilelayer" not in map.settings["properties"]


def test_can_replace_tilelayer_url_in_map_settings(map):
    map.settings["properties"]["tilelayer"] = {
        "name": "My TileLayer",
        "maxZoom": 18,
        "minZoom": 0,
        "attribution": "My attribution",
        "url_template": "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
    }
    map.save()
    new = "https://{s}.foo.bar.baz/{z}/{x}/{y}.png"
    call_command(
        "clean_tilelayer",
        "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
        new,
        "--no-input",
    )
    map = Map.objects.get(pk=map.pk)
    assert map.settings["properties"]["tilelayer"]["url_template"] == new


def test_can_replace_tilelayer_by_name_in_map_settings(map, tilelayer):
    map.settings["properties"]["tilelayer"] = {
        "name": "My TileLayer",
        "maxZoom": 18,
        "minZoom": 0,
        "attribution": "My attribution",
        "url_template": "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
    }
    map.save()
    call_command(
        "clean_tilelayer",
        "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
        tilelayer.name,
        "--no-input",
    )
    map = Map.objects.get(pk=map.pk)
    assert map.settings["properties"]["tilelayer"] == tilelayer.json


def test_can_replace_tilelayer_by_id_in_map_settings(map, tilelayer):
    map.settings["properties"]["tilelayer"] = {
        "name": "My TileLayer",
        "maxZoom": 18,
        "minZoom": 0,
        "attribution": "My attribution",
        "url_template": "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
    }
    map.save()
    call_command(
        "clean_tilelayer",
        "http://{s}.foo.bar.baz/{z}/{x}/{y}.png",
        tilelayer.pk,
        "--no-input",
    )
    map = Map.objects.get(pk=map.pk)
    assert map.settings["properties"]["tilelayer"] == tilelayer.json
